import argparse
import glob
import os
import sqlite3
from datetime import datetime, timezone
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from shutil import copyfileobj
from subprocess import check_output
from typing import Optional

import yaml
from jinja2 import Template

from xfel_calibrate.settings import free_nodes_cmd, preempt_nodes_cmd

try:
    from .config import serve_overview as config
except:
    from config import serve_overview as config


# HTTPRequestHandler class
class RequestHandler(BaseHTTPRequestHandler):
    conf_was_init = False

    def init_config(self):
        self.total_jobs_cmd = config["shell-commands"]["total-jobs"]
        self.tail_log_cmd = config["shell-commands"]["tail-log"]
        self.cat_log_cmd = config["shell-commands"]["cat-log"]
        self.run_candidates = config["run-candidates"]

        self.templates = {}
        for template, tfile in config["templates"].items():
            with open(tfile, "r") as tf:
                self.templates[template] = tf.read()

        self.conf_was_init = True

    def serve_css(self):
        """Serve /serve_overview.css"""
        self.send_response(200)
        self.send_header('Content-type', 'text/css')
        self.end_headers()
        self.wfile.write(self.templates["css"].encode('utf-8'))

    def serve_dark_overview(self):
        # Send headers
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        host = config["server-config"]["host"]
        port = config["server-config"]["port"]
        reports = {}
        for instrument, detectors in cal_config['dark'].items():
            reports[instrument] = {}
            for detector in detectors:
                tmpl = f'/gpfs/exfel/d/cal/caldb_store/xfel/reports/{instrument}/{detector}/dark/*pdf'
                files = glob.glob(tmpl)
                files.sort(key=os.path.getmtime, reverse=True)
                file_info = []
                for i, file in enumerate(files):
                    if (len(file_info) % 2) == 0:
                        bgcolor = 'EEEEEE'
                    else:
                        bgcolor = 'FFFFFF'
                    time = os.stat(file).st_mtime
                    d_time = datetime.fromtimestamp(time).replace(
                        tzinfo=timezone.utc)
                    s_time = d_time.strftime('%y-%m-%d %H:%M')
                    file_info.append([file, s_time, bgcolor])

                reports[instrument][detector] = file_info

        tmpl = Template(self.templates["dark-overview"])
        message = tmpl.render(reports=reports, host=host, port=port)

        self.wfile.write(bytes(message, "utf8"))

    def serve_file_from_gpfs(self):
        """Serve a file from a path starting with /gpfs"""
        if self.path.endswith(".html"):
            mimetype = 'text/html'
        elif self.path.endswith(".jpg"):
            mimetype = 'image/jpg'
        elif self.path.endswith(".gif"):
            mimetype = 'image/gif'
        elif self.path.endswith(".png"):
            mimetype = 'image/png'
        elif self.path.endswith(".pdf"):
            mimetype = 'application/pdf'
        elif self.path.endswith(".js"):
            mimetype = 'application/javascript'
        elif self.path.endswith(".css"):
            mimetype = 'text/css'
        else:
            return self.send_error(404)

        if os.path.isfile(self.path):
            self.send_response(200)
            self.send_header('Content-Length', str(os.stat(self.path).st_size))
            self.send_header('Content-type', mimetype)
            self.end_headers()
            with open(self.path, "rb") as f:
                copyfileobj(f, self.wfile)
        else:
            self.send_error(404)

    def do_GET(self):

        if not self.conf_was_init:
            self.init_config()

        if "/serve_overview.css" in self.path:
            return self.serve_css()

        if "dark?" in self.path:
            return self.serve_dark_overview()

        if "/gpfs" in self.path:
            return self.serve_file_from_gpfs()

        # Send response status code
        self.send_response(200)

        # Send headers
        self.send_header('Content-type', 'text/html')
        self.end_headers()

        # Send message back to client
        # general maxwell stats
        free = int(check_output(free_nodes_cmd, shell=True).decode('utf8'))
        preempt = int(check_output(
            preempt_nodes_cmd, shell=True).decode('utf8'))
        nodes_avail_general = free + preempt
        total_jobs_running = check_output(self.total_jobs_cmd, shell=True)
        total_jobs_running = total_jobs_running.decode('utf8').split("/")[0]

        tmpl = Template(self.templates["maxwell-status"])
        maxwell_status_r = tmpl.render(nodes_avail_general=nodes_avail_general,
                                       total_jobs_running=total_jobs_running,
                                      )

        last_n_lines = check_output(self.tail_log_cmd,
                                    shell=True).decode('utf8').split("\n")
        last_n_lines = [l for l in last_n_lines
                        if ("Response error from MDC" not in l
                            and "DEBUG" not in l)]
        tmpl = Template(self.templates["log-output"])
        log_output_r = tmpl.render(service="Webservice", lines=last_n_lines)

        last_n_lines_monitor = [l for l in check_output(
                config["shell-commands"]["tail-log-monitor"], shell=True
            ).decode('utf8').split("\n")
            if "DEBUG" not in l
        ]
        log_output_monitor_r = tmpl.render(
            service="Job monitor", lines=last_n_lines_monitor
        )

        last_n_lines = check_output(self.cat_log_cmd,
                                    shell=True).decode('utf8').split("\n")[
                       ::-1]

        def get_run_info(l, key):
            """
            Parse a line and extract information
            :param l: Line to parse
            :param key: A key work: DARK or CORRECT
            :return: Detector name, Instrument name, input folder,
            output folder, path to report, list of runs, time of request
            """
            if key in l:
                ls = l.split()
                # karabo-id and report-to are expected to be
                # in the log file
                if any(x not in ls for x in
                       [key, '--karabo-id', '--report-to']):
                    return None

                dclass = ls[ls.index(key) - 1]
                in_folder = ls[ls.index("--in-folder") + 1]
                report_to = ls[ls.index("--report-to") + 1]
                out_folder = ls[ls.index("--out-folder") + 1]
                detector = ls[ls.index("--karabo-id") + 1]
                instrument = in_folder.split('/')[4]

                runs = []
                for rc in self.run_candidates:
                    if rc in ls:
                        runs.append(ls[ls.index(rc) + 1])

                requested = "{} {}".format(ls[0], ls[1])

                return [dclass, detector, instrument, in_folder, out_folder,
                        report_to, runs, requested]

        last_chars = {}
        last_calib = {}
        host = config["server-config"]["host"]
        port = config["server-config"]["port"]
        for l in last_n_lines:
            # Extract dark requests
            info = get_run_info(l, 'DARK')
            if info is not None:
                dclass, detector, instrument, in_folder, out_folder, report_to, runs, requested = info  # noqa

                # Check if instrument is in detector name
                # This is not the case for e.g. CALLAB
                key = detector if instrument in detector else f"{instrument}-{detector}"  # noqa
                if key in last_chars:
                    continue

                pdfs = glob.glob(f"{report_to}*.pdf")
                pdfs.sort()
                pdfs = {p.split("/")[-1]: p for p in pdfs}
                fpdfs = []
                if len(pdfs):
                    for pdf, p in pdfs.items():
                        fpdfs.append(
                            (pdf, f"http://{host}:{port}/{p}"))
                pdfs = fpdfs
                tsize = 0
                for run in runs:
                    run = int(run)
                    if detector not in cal_config['data-mapping']:
                        continue
                    # ToDo calculate tsize based on selected karabo-da
                    for mp in cal_config['data-mapping'][detector]['karabo-da']:
                        for f in glob.glob(
                                f"{in_folder}/r{run:04d}/*{mp}*.h5"):
                            tsize += os.stat(f).st_size
                constant_valid_from = ""

                last_chars[key] = {"in_path": in_folder,
                                   "out_path": out_folder,
                                   "runs": runs,
                                   "pdfs": pdfs,
                                   "size": "{:0.1f} GB".format(tsize / 1e9),
                                   "requested": requested,
                                   "device_type": detector,
                                   "last_valid_from": constant_valid_from}

            # Extract correction requests
            info = get_run_info(l, 'CORRECT')
            if info is not None:
                _, _, _, in_folder, _, report_to, runs, requested = info
                instrument = in_folder.split('/')[4]
                if instrument not in last_calib:
                    last_calib[instrument] = []
                if len(last_calib[instrument]) > config["server-config"]["n-calib"]:   # noqa
                    continue
                proposal = in_folder.split('/')[6]
                pdfs = glob.glob(f"{report_to}*.pdf")
                pdfs.sort(key=os.path.getmtime, reverse=True)
                pdfs = {p.split("/")[-1]: p for p in pdfs}

                if [proposal, runs[0]] not in [[x[1], x[2]] for x in
                                               last_calib[instrument]]:
                    last_calib[instrument].append([requested[:-4],
                                                   proposal, runs[0], pdfs])

        tmpl = self.templates["last-characterizations"]
        last_characterizations_r = Template(tmpl).render(char_runs=last_chars,
                                                         host = host,
                                                         port = port)

        tmpl = self.templates["last-correction"]
        last_correction_r = Template(tmpl).render(info=last_calib, host=host,
                                                  port=port)

        conn = sqlite3.connect(config['web-service']['job-db']).cursor()
        conn.execute("SELECT * FROM jobs WHERE status IN ('R', 'PD', 'CG')")
        running_jobs = {}
        for r in conn.fetchall():
            rid, jobid, proposal, run, status, time, det, act = r
            run = int(run)
            key = '{}/r{:04d}/{}/{}'.format(proposal, run, det, act)
            flg = "R"
            if status in ["QUEUE", "PD"]:
                flg = "Q"
            rjobs = running_jobs.get(key, [])
            rjobs.append((flg, '{}-{}'.format(status, time)))
            running_jobs[key] = rjobs

        tmpl = self.templates["running-jobs"]
        running_jobs_r = Template(tmpl).render(running_jobs=running_jobs)

        tmpl = Template(self.templates["main-doc"])
        message = tmpl.render(maxwell_status=maxwell_status_r,
                              log_output=log_output_r,
                              log_output_monitor=log_output_monitor_r,
                              last_characterizations=last_characterizations_r,
                              last_correction=last_correction_r,
                              running_jobs=running_jobs_r)
        # Write content as utf-8 data
        self.wfile.write(bytes(message, "utf8"))
        return


def run(config_file: Optional[str] = None):
    if config_file is not None:
        config.configure(includes_for_dynaconf=[Path(config_file).absolute()])

    with open(config["web-service"]["cal-config"], "r") as cf:
        global cal_config
        cal_config = yaml.load(cf.read(), Loader=yaml.FullLoader)
    print('starting server...')
    sconfig = config["server-config"]
    server_address = (sconfig["host"], sconfig["port"])
    httpd = HTTPServer(server_address, RequestHandler)
    print('running server...')
    httpd.serve_forever()


parser = argparse.ArgumentParser(
    description='Start the overview server')
parser.add_argument('--config', type=str, default=None,)
if __name__ == "__main__":
    args = vars(parser.parse_args())
    run(args["config"])
